import random

class Canvas(object):
    """The canvas class manages a set of shapes and provides a coordinate system.
       Shapes may extend beyond the edge of the canvas, however their (x, y)
       coordinates will always be within the bounds of the canvas."""

    def __init__(self, width=200, height=200):
        """A canvas has a list of shapes and a width and a height.
           Coordinate (0,0) is located in the center of the canvas."""
        self.shapes = []
        self.width  = width
        self.height = height

    def add(self, shape):
        """Add a shape to the canvas."""
        self.shapes.append(shape)

    def alignHorizontally(self, shapes):
        """Align the given shapes with the Y coordinate of the first shape."""
        if len(shapes) > 0:
            targetY = shapes[0].getY()
            for s in shapes:
                s.setY(targetY)

    def alignVertically(self, shapes):
        """Align the given shapes with the X coordinate of the first shape."""
        if len(shapes) > 0:
            targetX = shapes[0].getX()
            for s in shapes:
                s.setX(targetX)
                
    def getRandomShape(self):
        return random.choice(self.shapes)

    def maxX(self):
        """Calculate the largest X coordinate supported by this canvas."""
        return self.width / 2

    def maxY(self):
        """Calculate the largest Y coordinate supported by this canvas."""
        return self.height / 2

    def minX(self):
        """Calculate the smallest X coordinate supported by this canvas."""
        return -self.width / 2

    def minY(self):
        """Calculate the smallest Y coordinate supported by this canvas."""
        return -self.height / 2

    def moveShapeTo(self, s, x, y):
        """Moves the given shape to a specific (x,y) coordinate. Ensures
           that the specified (x,y) coordinate is within the bounds of
           the canvas. The shape itself might extend beyond the boundaries
           of the canvas, but its (x,y) coordinate will always be within
           bounds."""
        if x < self.minX():
            x = self.minX()
        if x > self.maxX():
            x = self.maxX()
        if y < self.minY():
            y = self.minY()
        if y > self.maxY():
            y = self.maxY()
        s.setX(x)
        s.setY(y)
        
    def moveShapeBy(self, s, deltaX, deltaY):
        """Moves the given shape by a given x and y delta. Computes the
           new location and then uses self.moveShapeTo() to make
           sure the new location is within the bounds of the canvas."""
        x = s.getX()
        y = s.getY()
        x += deltaX
        y += deltaY
        self.moveShapeTo(s, x, y)

    def needS(self):
        """Utility function to make the word 'Shape' plural when needed..""" 
        if len(self.shapes) == 1:
            return ""
        else:
            return "s"

    def report(self):
        """Generate status report on each shaped contained in the canvas.""" 
        print "-" * 70
        print "Canvas contains %d Shape%s." % (len(self.shapes), self.needS())
        print
        for s in self.shapes:
            print "%s" % (s.report())
            print
        print "-" * 70

    def shapeWithMaxX(self):
        """Find the shape with the greatest X coordinate."""
        target  = None
        targetX = self.minX()
        for s in self.shapes:
            currentX = s.getX()
            if currentX > targetX:
                targetX = currentX
                target  = s
        return target

    def shapeWithMaxY(self):
        """Find the shape with the greatest Y coordinate."""
        target  = None
        targetY = self.minY()
        for s in self.shapes:
            currentY = s.getY()
            if currentY > targetY:
                targetY = currentY
                target  = s
        return target

    def shapeWithMinX(self):
        """Find the shape with the lowest X coordinate."""
        target  = None
        targetX = self.maxX()
        for s in self.shapes:
            currentX = s.getX()
            if currentX < targetX:
                targetX = currentX
                target  = s
        return target

    def shapeWithMinY(self):
        """Find the shape with the lowest Y coordinate."""
        target  = None
        targetY = self.maxY()
        for s in self.shapes:
            currentY = s.getY()
            if currentY < targetY:
                targetY = currentY
                target  = s
        return target